# Copyright (c) 2021 Edwin Wise
# MIT License
# See LICENSE for details
"""
    PiDeviceNet Module. Each device client should instantiate a module to
    publish the device data through, and to receive controller messages,
    and messages from other related modules
"""
import json
import logging
import socket
import struct
import time
from dataclasses import asdict
from uuid import uuid4

import paho.mqtt.client as mqtt

from pi_device_net.constants import (ModuleState, MULTICAST_GROUP,
                                     MULTICAST_PORT,
                                     DISCOVERY_PACKET_SIZE, Command,
                                     ControllerConnectReply, ModulePublish,
                                     ModuleFound)

LOG = logging.getLogger(__name__)


class ModuleNetwork():
    """ PiDeviceNet Module

    When you initialize you specify a message callback, which is how you
    will receive all messages.

    Your application must call `module.discover()` to connect to the
    Controller.
    """

    def __init__(self,
                 system_name,
                 module_name,
                 message_callback,
                 multicast_group=None,
                 multicast_port=None,
                 discovery_size=None):
        """ Initialize the Module.

        Parameters
        ----------
        system_name : String
            The name of the control network.  The controllers and all modules
            must be using the same system name.  You can run multiple isolated
            device networks on a single router if they have different system
            names.
        module_name : String
            A descriptive yet short name for this module.  Used in messaging.
        message_callback : Callable[String, String, Dict]
            For every system message received by the controller, this callable
            is called with (module_name, topic_string, message_dictionary)
        multicast_group : String
            The IP address of the multicast group used for module discovery.
            Defaults to a sensible broadcast address in `constants.py`
        multicast_port : Int
            The port that is used for the discovery socket. Defaults to the
            value in `constants.py`
        discovery_size : Int
            The size that the discovery messages are limited to.  Defaults
            to the value in `constants.py`
        """
        self.id = str(uuid4())
        self.system = system_name
        self.module_name = module_name
        self.host = socket.getfqdn()
        self.root_topic = None
        self.topics = set()
        self.state = ModuleState.STOPPED
        self.multicast_group = multicast_group or MULTICAST_GROUP
        self.multicast_port = multicast_port or MULTICAST_PORT
        self.discovery_size = discovery_size or DISCOVERY_PACKET_SIZE
        self._discovery_socket = None
        self._mqtt = None
        self._message_callback = message_callback

    def discover(self, timeout=None):
        """ Wait to be discovered by a controller.

        Connects to the controller when found. The default behavior is to
        lock up forever, opening and closing the discovery socket, until
        a controller is found.  This can be refined (for testing, if nothing
        else) by specifying `timeout` and `retries`.

        Parameters
        ----------
        timeout : Float
            How long the module tries to access the discovery port before
            giving up and closing the port.  Defaults to FOREVER.
        """
        self.state = ModuleState.DISCOVERY

        start_time = time.time()
        while self.state == ModuleState.DISCOVERY:
            # Make sure we have a discovery socket
            if self._discovery_socket is None:
                self._discovery_socket = self._connect_socket()

            try:
                data, address = self._discovery_socket.recvfrom(
                    self.discovery_size)
                try:
                    # Got data! See if it parses...
                    packet = json.loads(data)
                    system = packet.get('system')

                    # We only connect to the system we are a part of
                    if system == self.system:
                        command = packet.get('command')

                        if command == Command.DISCOVER.value:
                            # Reply that we have been found
                            self._reply_found(address)

                        elif command == Command.CONNECT.value:
                            # We have been told to connect, so do that
                            connect = ControllerConnectReply(**packet)

                            # But only if they are telling US to connect
                            if connect.id == self.id:
                                LOG.info(f"Connecting to {connect.broker} "
                                         f"as {connect.topic}")
                                self._mqtt = self._connect_mqtt(connect.broker,
                                                                connect.topic)

                                # We are running now, so the discovery loop
                                # will exit
                                self.state = ModuleState.RUNNING

                        else:
                            LOG.info(f"Unsupported discovery command "
                                     f"{command!r}")
                    else:
                        LOG.info(f"Unrecognized system {system!r} "
                                 f"at {address}")

                except Exception as e:
                    LOG.error(f"Problem reading incoming packet: "
                              f"{type(e)} {e}: \n{data}")
            except BlockingIOError as e:
                pass
            except Exception as e:
                LOG.error(f"Unmanaged exception {type(e)}: {e}")

            # No need to lock the CPU
            time.sleep(1)

            if timeout and (time.time() - start_time) > timeout:
                raise TimeoutError("Timed out waiting for discovery.")

    def stop(self):
        """ Stopping the module means disconnecting any discovery
        or messaging connections and changing state.
        """
        if self.state is ModuleState.STOPPED:
            LOG.warning("Already stopped.")
        if self._mqtt is not None:
            self._mqtt.disconnect()
            self._mqtt = None
        if self._discovery_socket is not None:
            self._discovery_socket.close()
            self._discovery_socket = None
        self.state = ModuleState.STOPPED

    def publish(self, topic, message):
        """ Publish a (module-specific) message.

        Parameters
        ----------
        topic : String
            The specific topic under this system/module to publish to:
                "<system_name>/<module_name>/<topic>"
        message : Any
            Any json-parsable message
        """
        if self.state is not ModuleState.RUNNING:
            LOG.warning("Attempting to send message before running.")

        cmd = ModulePublish(system=self.system,
                            module=self.module_name,
                            id=self.id,
                            message=message)
        packet = asdict(cmd)
        # QOS 1 : At least once
        self._mqtt.publish(f"{self.root_topic}/{topic}",
                           json.dumps(packet).encode(),
                           qos=1)

    def subscribe(self, topic, module=None):
        """ Subscribe to a topic.

        The most likely use is to get information from other modules
        so that this module can coordinate with another module (e.g.
        a flow totalizer reporting flow, that a flow controller can use
        to regulate the flow).

        Parameters
        ----------
        topic : String
            Specific topic to subscribe to
        module : String
            If specified, the module we want to subscribe against.  Defaults
            to the wildcard '+' to subscribe to all modules, giving a full
            topic path of:

                `<system_name>/<module>/<topic>`
            or:
                `<system_name>/+/<topic>`
        """
        if self.state is not ModuleState.RUNNING:
            LOG.warning("Attempting to subscribe before running.")

        module = module or '+'
        topic_path = f"{self.system}/{module}/{topic}"
        self.topics.add(topic_path)
        self._mqtt.subscribe(topic_path)

    def _connect_socket(self):
        """ Connect to the discovery socket.

        Returns
        -------
        Socket
        """
        sock = socket.socket(socket.AF_INET,
                             socket.SOCK_DGRAM)
        sock.setblocking(False)
        sock.bind(('', self.multicast_port))
        mreq = struct.pack('4sL',
                           socket.inet_aton(self.multicast_group),
                           socket.INADDR_ANY)
        sock.setsockopt(socket.IPPROTO_IP,
                        socket.IP_ADD_MEMBERSHIP,
                        mreq)
        return sock

    def _reply_found(self, address):
        """ Reply to a discovery broadcast, that we are found.

        address : (String, Int)
            Controller address as returned from the socket, as a tuple
            (<host>, <port>)
        """
        reply = ModuleFound(module=self.module_name,
                            id=self.id,
                            host=self.host)
        packet = asdict(reply)
        self._discovery_socket.sendto(json.dumps(packet).encode(), address)

    def _connect_mqtt(self, broker, topic):
        """ Connect to the messaging broker and set the topic for this
        module as set by the controller.

        Subscribes to "<topic>/admin" (which ends up being
        "<system_name>/<module>/admin") so the controller can talk directly
        to this module.

        Parameters
        ----------
        broker : String
            Broker IP address, sent as part of the Connect message
        topic : String
            Root topic in the form "<system_name>/<module>" as sent by
            the Connect message.

        Returns
        -------
        mqtt.Client
        """
        client = None
        if broker and topic:
            if self._discovery_socket is not None:
                self._discovery_socket.close()
                self._discovery_socket = None

            self.root_topic = topic
            admin_topic = f'{topic}/admin'
            self.topics.add(admin_topic)

            client = mqtt.Client(client_id=self.id)
            client.on_connect = self._connect_cb
            client.on_disconnect = self._disconnect_cb
            client.on_message = self._message_cb
            client.enable_logger()
            client.connect(broker)

            client.loop_start()
        return client

    def _connect_cb(self, client, userdata, flags, rc):
        """ MQTT callback on_connect; re-subscribes to all topics.
        """
        LOG.info("MQTT Connected")
        for topic in self.topics:
            LOG.info(f"... subscribing to {topic!r}")
            client.subscribe(topic)

    def _disconnect_cb(self, client, topic, flags, rc=0):
        """ MQTT callback on_disconnect
        """
        LOG.warning("MQTT Disconnected")

    def _message_cb(self, client, userdata, msg):
        """ MQTT callback on_message

        Parameters
        ----------
        msg : Any
            The message is any json-parsable entity
        """
        try:
            packet = json.loads(msg.payload)
        except Exception as e:
            return

        # Only manage messages to our system
        system = packet.get('system')
        if system != self.system:
            return

        try:
            # This system operates on three levels for the topic:
            # System name, Module name, Topic
            _, module, topic = msg.topic.split('/')
        except Exception as e:
            LOG.error(f"Ill formatted topic {msg.topic!r};"
                      f"{type(e)}: {e}")
            return

        self._message_callback(module, topic, packet.get('message'))



# Copyright (c) 2021 Edwin Wise
# MIT License
# See LICENSE for details
"""
    PiDeviceNet interfaces (and most of the Device implementation.
"""
import logging
import threading
from abc import ABC, abstractmethod

import pygame

from constants import SAMPLE_RATE

LOG = logging.getLogger(__name__)


class IDevice(ABC):
    """ Base class for devices, giving threaded behavior to query hardware.

    There are two sides to a Device.

    Device Thread:
        The child of IDevice implements the `_read()` method .

        When the client calls `run()`, a thread is started that repeatedly
        calls the specific `_read()` and stores the result.

    Device Client:
        To start the device thread(s), call `run()`.

        To see the most recent value, check the `value` property. If an
        archiver is specified, see the most recent set of values using
        the `history()` method.

    The device also supports two optional systems:

    * IStatistics object, to perform online statistical interpretation of the
        value stream

    * IArchiver object, to provide a memory for the value stream
    """
    def __init__(self, device_name,
                 statistics=None,
                 archiver=None):
        """ Initialize the device.

        Parameters
        ----------
        device_name : String
            The name of the device, used as part of the storage scheme and
            for other references.
        statistics : IStatistics
            If specified, each value will be sent to `update()` to track
            whatever details about the device's behavior the IStatistics
            tracks (`mean()` at a minimum).
        archiver : IArchiver
            If specified, each value will be sent to 'put()'. You can query
            the archiver using IDevice `history()` method or methods on the
            archiver itself.
        """
        self.name = device_name
        self.sample_rate = None

        self._statistics = statistics
        self._archiver = archiver
        self._thread = None
        self._running = False
        self._last_reading = None

    # === Data access

    @abstractmethod
    def _read(self):
        """ Read a value from the hardware. The only abstract method in this
        interface. The child class must implement the actual device read.

        Returns
        -------
        Number
            The value that represents the current state of the device. In
            theory it doesn't have to be a number, but some of the companion
            interfaces may expect a number.
        """

    @property
    def value(self):
        """ The last read value for this device.

        Returns
        -------
        Number (or whatever this device uses as its state)
        """
        return self._last_reading

    def list_history(self, tail=None):
        """ If an archiver was specified, returns the most recent set of
        values archived.

        Parameters
        ----------
        tail : Int
            If specified, determines how many values to return (clamped to
            the buffer size).  Defaults to returning all values in memory.

        Returns
        -------
        list(Number)
            The list may be empty or have fewer entries than requested,
            depending on the state of the archiver.
        """
        if self._archiver is None:
            return []
        return self._archiver.history(tail)

    # === Execution control

    def run(self, sample_rate=None):
        """ Start the device thread, which will call `_read()` at the
        specified sample rate.

        Parameters
        ----------
        sample_rate : Int
            The time between samples in milliseconds. Defaults to 1,000 (one
            sample per second).
        """
        if self._thread is not None:
            LOG.warning("Already running")
            return

        if self._archiver is not None:
            LOG.info("Starting archiver")
            self._archiver.run()

        self.sample_rate = sample_rate or SAMPLE_RATE
        self._thread = threading.Thread(target=self._run)
        self._thread.start()

    def stop(self):
        if self._thread is None:
            LOG.warning("Already stopped")
            return

        if self._archiver is not None:
            LOG.info("Stopping archiver")
            self._archiver.stop()

        self._running = False
        self._thread.join()
        self._thread = None

    # === Utility methods

    def map_ranges(self, input, input_min, input_max,
                   output_min, output_max):
        """ Utility method to map an input between two value ranges.

        Parameters
        ----------
        input : Number
            The number we are mapping to a new range
        input_min : Number
            The minimum value that `input` can take
        input_max : Number
            The maximum value that `input` can take
        output_min : Number
            The minimum value of the new output range
        output_max : Number
            The maximum value of the new output range

        Returns
        -------
        Number
            The input value as mapped to the output range
        """
        input_range = input_max - input_min
        output_range = output_max - output_min
        normalized_value = float(input - input_min) / float(input_range)
        return output_min + (normalized_value * output_range)

    # === Internal methods

    def _run(self):
        """ The threaded run method; reads the device at the sample rate,
        updates the device's last reading, and updates the statistics and
        archiver (if specified).

        Runs until stopped.
        """
        self._running = True
        while self._running:
            # Use the pygame framerate control
            pygame.time.wait(self.sample_rate)

            self._last_reading = self._read()
            if self._statistics:
                self._statistics.update(self._last_reading)
            if self._archiver is not None:
                self._archiver.put(self._last_reading)


class IStatistics(ABC):
    """ Online statistics, for tracking the behavior of an IDevice.
    """

    @abstractmethod
    def reset(self):
        """ Zero out the statistics held so far.
        """

    @abstractmethod
    def update(self, sample):
        """ Add a sample to the ongoing statistics.

        Parameters
        ----------
        sample : Number
            A new sample value to process. Does not have to be a number,
            if the device ecosystem as a whole is not using numbers.
        """

    @abstractmethod
    def mean(self):
        """ Retrieve the mean value of the samples to date.

        Returns
        -------
        Number
            The mean value (more or less) of all updates taken to date.
        """


class IArchiver(ABC):
    """ Manage values history for an IDevice.
    """

    @abstractmethod
    def run(self):
        """ Run the archiver, if it needs to start a thread. The archiver
        may use a separate thread or process to do slow activities, such
        as writing a chunk of the history to disk.
        """

    @abstractmethod
    def stop(self):
        """ Stop the archiver, if it started a thread.
        """

    @abstractmethod
    def put(self, value):
        """ Add a value to the archive.

        Parameters
        ----------
        value : Number
            The value to store in the archive.  Could technically be anything,
            but some of the other companion interfaces assume a numerical
            value.
        """

    @abstractmethod
    @property
    def history(self, tail=None):
        """ Return the history with an optional tail size.

        Parameters
        ----------
        tail : Int
            If specified, the number of values at the end of the archive to
            return; otherwise returns the entire archive in memory.

        Returns
        -------
        tuple(Number)
        """



# Copyright (c) 2021 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Manual test for controller/module discovery and messaging.

    To set up, the controller must be running mosquito:

        `sudo apt-get install mosquitto`

    Deploy this file and the device net source to the controller and
    module hardware.

    From the PiDeviceNet directory on the respective controller and module
    devices...

    Start the controller first with:

        `python3 -m pi_device_net.tests.manual_test_discovery controller`

    Then start the module with:

        `python3 -m pi_device_net.tests.manual_test_discovery module`

    They will both log some messages and the module will send ten updates
    to the controller, after which they will both exit.

    If you start the module before the controller, they should still find
    each other and connect.
"""
import json
import logging
import sys
import time

from pi_device_net.controller import ControlNetwork
from pi_device_net.module import ModuleNetwork

SYSTEM = "testing"

logging.basicConfig(
    level=logging.INFO,
    format="[%(asctime)s] [%(levelname)s] "
           "[%(name)s.%(lineno)d] %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S %z")
LOG = logging.getLogger('test_discovery')

if __name__ == '__main__':
    if len(sys.argv) < 2:
        print(f"Need to run as either 'controller' or 'module'.")
        exit(1)

    mode = sys.argv[1]

    # We only want to process 10 messages before stopping the controller
    counter = 0

    def on_controller_message(module, topic, packet):
        global counter
        print(f"CONTROLLER: module {module!r} : "
              f"topic {topic!r} : "
              f"{json.dumps(packet, sort_keys=True)}")
        counter += 1

    def on_module_message(module, topic, packet):
        print(f"MODULE: module {module!r} : "
              f"topic {topic!r} : "
              f"{json.dumps(packet, sort_keys=True)}")

    if mode == 'controller':
        controller = ControlNetwork(SYSTEM, on_controller_message)
        while counter < 10:
            # Broadcast a discovery request (and connect to any modules
            # responding to it)
            controller.discover()

            # For every module we know about...
            modules = controller.list_modules()
            for module in modules:
                # ... send them an administration message test
                controller.publish(f"{module}/admin", 'ping')

            # Be kind to our CPU and logging output
            time.sleep(1)

    elif mode == 'module':
        module = ModuleNetwork(SYSTEM, 'module', on_module_message)

        # Block until we get and process a discovery request from
        # a controller (at least, until we give up).
        module.discover()
        for i in range(10):
            module.publish('counter', {'count': i})
            time.sleep(1)
        module.stop()

    else:
        print(f"Unsupported mode: {mode!r}; must be 'head' or 'tail'.")
        exit(1)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pi-device-net",
    version="0.1.9",
    author="Edwin Wise",
    author_email="edwin@simreal.com",
    description="A set of tools to operate a device network.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/EdwinWiseOne/PiDeviceNet",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: Free for non-commercial use",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Home Automation",
        "Topic :: Software Development :: Libraries",
    ],
    python_requires='>=3.7',
    install_requires=['paho-mqtt', 'pygame'],
)